from collections import defaultdict
from wsgiref.util import request_uri
import tqdm
import pandas as pd
import pysam
from Bio import SeqIO, Seq, SeqRecord
import logging


log = logging.getLogger("my_logger")


def get_read_by_taxa(samfile, refs_tax, refs, refs_damaged):

    reads = defaultdict(lambda: defaultdict(dict))

    for reference in tqdm.tqdm(
        refs,
        total=len(refs),
        leave=False,
        desc="References processed",
    ):
        for aln in tqdm.tqdm(
            samfile.fetch(reference=reference, multiple_iterators=False),
            total=samfile.count(reference=reference),
            ncols=80,
            ascii="░▒█",
            leave=False,
            desc="Alignments processed",
        ):
            # create read
            # Check if reference is damaged
            if aln.reference_name in refs_damaged:
                is_damaged = "damaged"
            else:
                is_damaged = "non-damaged"
            if reads[refs_tax[aln.reference_name]][aln.qname]:
                dmg = reads[refs_tax[aln.reference_name]][aln.qname]["is_damaged"]
                if dmg == is_damaged:
                    continue
                else:
                    reads[refs_tax[aln.reference_name]][aln.qname][
                        "is_damaged"
                    ] = "multi"
            else:
                seq = Seq.Seq(aln.seq)
                qual = aln.query_qualities
                if aln.is_reverse:
                    seq = seq.reverse_complement()
                    qual = qual[::-1]
                reads[refs_tax[aln.reference_name]][aln.qname] = {
                    "seq": seq,
                    "qual": qual,
                    "is_damaged": is_damaged,
                }

    samfile.close()
    return reads



import argparse
import sys
import gzip
import os
import logging
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull
from dmg_reads import __version__
import time
import json
from dmg_reads.defaults import mdmg_header, valid_ranks
from collections import defaultdict
import re
from itertools import chain

log = logging.getLogger("my_logger")
log.setLevel(logging.INFO)
timestr = time.strftime("%Y%m%d-%H%M%S")


def is_debug():
    return logging.getLogger("my_logger").getEffectiveLevel() == logging.DEBUG


# From: https://note.nkmk.me/en/python-check-int-float/
def is_integer(n):
    try:
        float(n)
    except ValueError:
        return False
    else:
        return float(n).is_integer()


# function to check if the input value has K, M or G suffix in it
def check_suffix(val, parser, var):
    if var == "--scale":
        units = ["K", "M"]
    else:
        units = ["K", "M", "G"]
    unit = val[-1]
    value = int(val[:-1])

    if is_integer(value) & (unit in units) & (value > 0):
        if var == "--scale":
            if unit == "K":
                val = value * 1000
            elif unit == "M":
                val = value * 1000000
            elif unit == "G":
                val = value * 1000000000
            return val
        else:
            return val
    else:
        parser.error(
            "argument %s: Invalid value %s. Has to be an integer larger than 0 with the following suffix K, M or G"
            % (var, val)
        )


# From https://stackoverflow.com/a/59617044/15704171
def convert_list_to_str(lst):
    n = len(lst)
    if not n:
        return ""
    if n == 1:
        return lst[0]
    return ", ".join(lst[:-1]) + f" or {lst[-1]}"


def get_compression_type(filename):
    """
    Attempts to guess the compression (if any) on a file using the first few bytes.
    http://stackoverflow.com/questions/13044562
    """
    magic_dict = {
        "gz": (b"\x1f", b"\x8b", b"\x08"),
        "bz2": (b"\x42", b"\x5a", b"\x68"),
        "zip": (b"\x50", b"\x4b", b"\x03", b"\x04"),
    }
    max_len = max(len(x) for x in magic_dict)

    unknown_file = open(filename, "rb")
    file_start = unknown_file.read(max_len)
    unknown_file.close()
    compression_type = "plain"
    for file_type, magic_bytes in magic_dict.items():
        if file_start.startswith(magic_bytes):
            compression_type = file_type
    if compression_type == "bz2":
        sys.exit("Error: cannot use bzip2 format - use gzip instead")
        sys.exit("Error: cannot use zip format - use gzip instead")
    return compression_type


def get_open_func(filename):
    if get_compression_type(filename) == "gz":
        return gzip.open
    else:  # plain text
        return open


def check_values(val, minval, maxval, parser, var):
    value = float(val)
    if value < minval or value > maxval:
        parser.error(
            "argument %s: Invalid value %s. Range has to be between %s and %s!"
            % (
                var,
                value,
                minval,
                maxval,
            )
        )
    return value


# From: https://stackoverflow.com/a/11541450
def is_valid_file(parser, arg, var):
    if not os.path.exists(arg):
        parser.error("argument %s: The file %s does not exist!" % (var, arg))
    else:
        return arg


def is_valid_filter(parser, arg, var):
    arg = json.loads(arg)
    # check if the dictionary keys are in the mdmg header list
    for key in arg.keys():
        if key not in mdmg_header:
            parser.error(
                f"argument {var}: Invalid value {key}.\n"
                f"Valid values are: {convert_list_to_str(mdmg_header)}"
            )

    return arg


def is_valid_rank(parser, arg, var):
    arg = json.loads(arg)
    for key in arg.keys():
        if key not in valid_ranks.keys():
            parser.error(
                f"argument {var}: Invalid value {key}.\n"
                f"Valid values are: {convert_list_to_str(valid_ranks)}"
            )
    return arg


def get_ranks(parser, ranks, var):
    valid_ranks = [
        "domain",
        "kingdom",
        "lineage",
        "phylum",
        "class",
        "order",
        "family",
        "genus",
        "species",
    ]
    ranks = ranks.split(",")
    # check if ranks are valid
    for rank in ranks:
        if rank not in valid_ranks:
            parser.error(
                f"argument {var}: Invalid value {rank}.\Rank has to be one of {convert_list_to_str(valid_ranks)}"
            )
        if rank == "all":
            ranks = valid_ranks[1:]
    return ranks


defaults = {
    "metaDMG_filter": {"Bayesian_D_max": 0.1, "Bayesian_z": 2.5},
    "prefix": None,
    "sort_memory": "1G",
    "threads": 1,
}

help_msg = {
    "metaDMG_results": "A file from metaDMG ran in local mode",
    "metaDMG_filter": "Which filter to use for metaDMG results",
    "bam": "The BAM file used to generate the metaDMG results",
    "prefix": "Prefix used for the output files",
    "taxonomy_file": "A file containing the taxonomy of the BAM references in the format d__;p__;c__;o__;f__;g__;s__.",
    "rank": "Which taxonomic group and rank we want to get the reads extracted.",
    "combine": "If set, the reads damaged and non-damaged will be combined in one fastq file",
    "only_damaged": "If set, only the reads damaged will be extracted",
    "sort_memory": "Set maximum memory per thread for sorting; suffix K/M/G recognized",
    "threads": "Number of threads",
    "debug": "Print debug messages",
    "version": "Print program version",
}


def get_arguments(argv=None):
    parser = argparse.ArgumentParser(
        description="A simple tool to extract damaged reads from BAM files",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-m",
        "--metaDMG-results",
        type=lambda x: is_valid_file(parser, x, "--metaDMG-results"),
        dest="metaDMG_results",
        help=help_msg["metaDMG_results"],
        required=True,
    )
    parser.add_argument(
        "-f",
        "--metaDMG-filter",
        type=lambda x: is_valid_filter(parser, x, "--metaDMG-filter"),
        dest="metaDMG_filter",
        help=help_msg["metaDMG_filter"],
        required=True,
    )
    parser.add_argument(
        "-b",
        "--bam",
        type=lambda x: is_valid_file(parser, x, "--bam"),
        dest="bam",
        help=help_msg["bam"],
    )
    parser.add_argument(
        "-p",
        "--prefix",
        type=str,
        default=defaults["prefix"],
        dest="prefix",
        help=help_msg["prefix"],
    )
    parser.add_argument(
        "--combine", dest="combine", action="store_true", help=help_msg["combine"]
    )
    parser.add_argument(
        "--only-damaged",
        dest="only_damaged",
        action="store_true",
        help=help_msg["only_damaged"],
    )
    parser.add_argument(
        "-T",
        "--taxonomy-file",
        type=lambda x: is_valid_file(parser, x, "---taxonomy-file"),
        dest="taxonomy_file",
        help=help_msg["taxonomy_file"],
    )
    parser.add_argument(
        "-r",
        "--rank",
        type=lambda x: is_valid_rank(parser, x, "--rank"),
        dest="rank",
        help=help_msg["rank"],
    )
    parser.add_argument(
        "-M",
        "--sort-memory",
        type=lambda x: check_suffix(x, parser=parser, var="--sort-memory"),
        default=defaults["sort_memory"],
        dest="sort_memory",
        help=help_msg["sort_memory"],
    )
    parser.add_argument(
        "-t",
        "--threads",
        type=lambda x: int(
            check_values(x, minval=1, maxval=1000, parser=parser, var="--threads")
        ),
        dest="threads",
        default=1,
        help=help_msg["threads"],
    )
    parser.add_argument(
        "--debug", dest="debug", action="store_true", help=help_msg["debug"]
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
        help=help_msg["version"],
    )
    args = parser.parse_args(None if sys.argv[1:] else ["-h"])
    return args


@contextmanager
def suppress_stdout():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


def create_output_files(prefix, bam, taxon=None, combined=False):
    if prefix is None:
        prefix = bam.split(".")[0]
    # create output files
    if taxon is None:
        if combined:
            out_files = {
                "fastq_combined": f"{prefix}.fastq.gz",
            }
        else:
            out_files = {
                "fastq_damaged": f"{prefix}.damaged.fastq.gz",
                "fastq_nondamaged": f"{prefix}.non-damaged.fastq.gz",
                "fastq_multi": f"{prefix}.multi.fastq.gz",
                "fastq_combined": f"{prefix}.fastq.gz",
            }
    else:
        if combined:
            out_files = defaultdict()
            for k, v in taxon.items():
                v = re.sub("[^0-9a-zA-Z]+", "_", v)
                out_files[f"fastq_{k}{v}"] = f"{prefix}.{k}{v}.fastq.gz"
        else:
            out_files = defaultdict()
            for k, v in taxon.items():
                v = re.sub("[^0-9a-zA-Z]+", "_", v)
                out_files[f"fastq_damaged_{k}{v}"] = f"{prefix}.{k}{v}.damaged.fastq.gz"
                out_files[
                    f"fastq_nondamaged_{k}{v}"
                ] = f"{prefix}.{k}{v}.non-damaged.fastq.gz"
                out_files[f"fastq_multi_{k}{v}"] = f"{prefix}.{k}{v}.multi.fastq.gz"
                out_files[f"fastq_combined_{k}{v}"] = f"{prefix}.{k}{v}.fastq.gz"
    return out_files


# From https://stackoverflow.com/a/61436083
def splitkeep(s, delimiter):
    split = s.split(delimiter)
    return [substr + delimiter for substr in split[:-1]] + [split[-1]]


def fast_flatten(input_list):
    return list(chain.from_iterable(input_list))



mdmg_header = [
    "tax_id",
    "sample",
    "N_reads",
    "lambda_LR",
    "D_max",
    "mean_L",
    "mean_GC",
    "q",
    "A",
    "c",
    "phi",
    "rho_Ac",
    "valid",
    "asymmetry",
    "std_L",
    "std_GC",
    "D_max_std",
    "q_std",
    "phi_std",
    "A_std",
    "c_std",
    "lambda_LR_P",
    "lambda_LR_z",
    "forward_lambda_LR",
    "forward_D_max",
    "forward_D_max_std",
    "forward_q",
    "forward_q_std",
    "forward_phi",
    "forward_phi_std",
    "forward_A",
    "forward_A_std",
    "forward_c",
    "forward_c_std",
    "forward_rho_Ac",
    "forward_lambda_LR_P",
    "forward_lambda_LR_z",
    "forward_valid",
    "reverse_lambda_LR",
    "reverse_D_max",
    "reverse_D_max_std",
    "reverse_q",
    "reverse_q_std",
    "reverse_phi",
    "reverse_phi_std",
    "reverse_A",
    "reverse_A_std",
    "reverse_c",
    "reverse_c_std",
    "reverse_rho_Ac",
    "reverse_lambda_LR_P",
    "reverse_lambda_LR_z",
    "reverse_valid",
    "LR_All",
    "LR_ForRev",
    "LR_ForRev_All",
    "chi2_all",
    "chi2_forward",
    "chi2_reverse",
    "chi2_ForRev",
    "N_x=1_forward",
    "N_x=1_reverse",
    "N_sum_total",
    "N_sum_forward",
    "N_sum_reverse",
    "N_min",
    "k_sum_total",
    "k_sum_forward",
    "k_sum_reverse",
    "Bayesian_z",
    "Bayesian_D_max",
    "Bayesian_D_max_std",
    "Bayesian_A",
    "Bayesian_A_std",
    "Bayesian_q",
    "Bayesian_q_std",
    "Bayesian_c",
    "Bayesian_c_std",
    "Bayesian_phi",
    "Bayesian_phi_std",
    "Bayesian_rho_Ac",
    "Bayesian_significance",
]

valid_ranks = {
    "domain": "d__",
    "kingdom": "k__",
    "lineage": "l__",
    "phylum": "p__",
    "class": "c__",
    "order": "o__",
    "family": "f__",
    "genus": "g__",
    "species": "s__",
}


"""
 Copyright (c) 2022 Antonio Fernandez-Guerra

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 the Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 """


import logging

from dmg_reads.utils import get_arguments, create_output_files, splitkeep, fast_flatten
from dmg_reads.lib import load_mdmg_results, filter_damaged_taxa
from dmg_reads.defaults import valid_ranks
from dmg_reads.extract import get_read_by_taxa
import pandas as pd
import os
import pysam
import numpy as np
from Bio import SeqIO, Seq, SeqRecord
import gzip
from mimetypes import guess_type
from functools import partial
import tqdm
from collections import defaultdict
import re

log = logging.getLogger("my_logger")


def main():

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)s ::: %(asctime)s ::: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    args = get_arguments()
    logging.getLogger("my_logger").setLevel(
        logging.DEBUG if args.debug else logging.INFO
    )

    # Check that rank and taxonomy file are both present
    if args.rank and not args.taxonomy_file:
        log.error("--rank requires --taxonomy")
        exit(1)
    if args.taxonomy_file and not args.rank:
        log("Error: --taxonomy-file requires --rank")
        exit(1)

    log.info("Loading metaDMG results...")
    mdmg_results = load_mdmg_results(args.metaDMG_results)
    # find which taxon are damaged

    damaged_taxa = filter_damaged_taxa(
        df=mdmg_results,
        filter_conditions=args.metaDMG_filter,
    )

    if args.taxonomy_file:
        log.info("Loading taxonomy data...")
        refs_tax = defaultdict()
        taxonomy = pd.read_csv(
            args.taxonomy_file,
            sep="\t",
            header=None,
            names=["reference", "taxonomy"],
        )

        # get get_ranks
        ranks = {valid_ranks[x]: args.rank[x] for x in args.rank if x in valid_ranks}
        # get refs that belong to this taxonomy
        # loop over all rows
        for i, row in tqdm.tqdm(
            taxonomy.iterrows(),
            total=taxonomy.shape[0],
            desc="Taxa processed",
            leave=False,
        ):
            taxs = row[1].split(";")
            for r in taxs:
                r1 = splitkeep(r, "__")
                if r1[0] in ranks and ranks[r1[0]] == r1[1]:
                    v = re.sub("[^0-9a-zA-Z]+", "_", r1[1])
                    refs_tax[row[0]] = f"{r1[0]}{v}"
                    # refs[row[0]] = f
    else:
        ranks = None

    logging.info("Loading BAM file...")
    save = pysam.set_verbosity(0)
    bam = args.bam
    samfile = pysam.AlignmentFile(bam, "rb")

    chr_lengths = []
    for chrom in samfile.references:
        chr_lengths.append(samfile.get_reference_length(chrom))
    max_chr_length = np.max(chr_lengths)

    # Check if BAM files is not sorted by coordinates, sort it by coordinates
    if not samfile.header["HD"]["SO"] == "coordinate":
        log.info("BAM file is not sorted by coordinates, sorting it...")
        sorted_bam = bam.replace(".bam", ".dr-sorted.bam")
        pysam.sort(
            "-@", str(args.threads), "-m", str(args.sort_memory), "-o", sorted_bam, bam
        )
        bam = sorted_bam
        samfile = pysam.AlignmentFile(bam, "rb")

    if not samfile.has_index():
        logging.info("BAM index not found. Indexing...")
        if max_chr_length > 536870912:
            logging.info("A reference is longer than 2^29, indexing with csi")
            pysam.index(bam, "-c")
        else:
            pysam.index(bam)  # Need to reload the samfile after creating index
            log.info("Re-loading BAM file")
            samfile = pysam.AlignmentFile(bam, "rb")
    pysam.set_verbosity(save)

    refs_bam = [
        chrom.contig for chrom in samfile.get_index_statistics() if chrom.mapped > 0
    ]

    if args.taxonomy_file:
        refs_damaged = set(refs_tax.keys()).intersection(
            set(damaged_taxa["reference"].to_list())
        )
        refs_non_damaged = (
            set(refs_tax.keys()).intersection(set(refs_bam)) - refs_damaged
        )
        if not refs_damaged and not refs_non_damaged:
            log.info("No references found in BAM file")
            exit(0)
    else:
        refs_damaged = damaged_taxa["reference"].to_list()
        refs_non_damaged = set(refs_bam) - set(refs_damaged)
        refs_tax = {ref: "all" for ref in refs_bam}

    if args.only_damaged:
        refs = refs_damaged
    else:
        refs = fast_flatten([refs_non_damaged, refs_damaged])

    out_files = create_output_files(prefix=args.prefix, bam=args.bam, taxon=ranks)

    for file in out_files:
        # file exists deleted
        if os.path.exists(out_files[file]):
            os.remove(out_files[file])

    log.info("Processing reads...")

    reads = get_read_by_taxa(
        samfile=samfile, refs=refs, refs_tax=refs_tax, refs_damaged=refs_damaged
    )
    # write reads

    count = defaultdict(int)
    logging.info("Saving reads...")
    if args.taxonomy_file:
        desc = "Taxa processed"
    else:
        desc = "References processed"

    for tax in tqdm.tqdm(reads, ncols=80, desc=desc, leave=False, total=len(reads)):
        if args.taxonomy_file:
            fastq_damaged = out_files[f"fastq_damaged_{tax}"]
            fastq_nondamaged = out_files[f"fastq_nondamaged_{tax}"]
            fastq_multi = out_files[f"fastq_multi_{tax}"]
            fastq_combined = out_files[f"fastq_combined_{tax}"]
        else:
            fastq_damaged = out_files["fastq_damaged"]
            fastq_nondamaged = out_files["fastq_nondamaged"]
            fastq_multi = out_files["fastq_multi"]
            fastq_combined = out_files["fastq_combined"]

        encoding = guess_type(fastq_damaged)[1]
        _open = partial(gzip.open, mode="at") if encoding == "gzip" else open
        # TODO: clean this up
        with _open(fastq_damaged) as f_damaged, _open(
            fastq_nondamaged
        ) as f_nondamaged, _open(fastq_multi) as f_multi, _open(
            fastq_combined
        ) as f_combined:
            for read in tqdm.tqdm(
                reads[tax],
                ncols=80,
                desc="Reads written",
                leave=False,
                total=len(reads[tax]),
                ascii="░▒█",
            ):
                rec = SeqRecord.SeqRecord(reads[tax][read]["seq"], read, "", "")
                rec.letter_annotations["phred_quality"] = reads[tax][read]["qual"]
                if args.combine:
                    SeqIO.write(rec, f_combined, "fastq")
                    if args.taxonomy_file:
                        count[out_files[f"fastq_combined_{tax}"]] += 1
                    else:
                        count[out_files["fastq_combined"]] += 1
                else:
                    if reads[tax][read]["is_damaged"] == "damaged":
                        SeqIO.write(rec, f_damaged, "fastq")
                        if args.taxonomy_file:
                            count[out_files[f"fastq_damaged_{tax}"]] += 1
                        else:
                            count[out_files["fastq_damaged"]] += 1
                    elif reads[tax][read]["is_damaged"] == "non-damaged":
                        SeqIO.write(rec, f_nondamaged, "fastq")
                        if args.taxonomy_file:
                            count[out_files[f"fastq_nondamaged_{tax}"]] += 1
                        else:
                            count[out_files["fastq_nondamaged"]] += 1
                    elif reads[tax][read]["is_damaged"] == "multi":
                        SeqIO.write(rec, f_multi, "fastq")
                        if args.taxonomy_file:
                            count[out_files[f"fastq_multi_{tax}"]] += 1
                        else:
                            count[out_files["fastq_multi"]] += 1

    for file in out_files:
        if count[out_files[file]] and count[out_files[file]] > 0:
            pass
        else:
            os.remove(out_files[file])

    logging.info("Done!")


if __name__ == "__main__":
    main()



import pandas as pd


def filter_damaged_taxa(df, filter_conditions):
    """Function to filter damaged taxa
    Args:
        df (panda.DataFrame): A dataframe containing metaDMG results
        filter_conditions (dict): A dictionary with filter conditions
        taxonomic_rank (str): Select the taxonomic rank to filter
    Returns:
        pandas.DataFrame: A filtered dataframe containing metaDMG results
    """

    mdmg_results = df.loc[
        (df[list(filter_conditions)] >= pd.Series(filter_conditions)).all(axis=1)
    ].copy()

    return mdmg_results


def load_mdmg_results(file_path):
    """Function to read a mdmg results file to a pandas dataframe
    Args:
        file_path (str): A file path pointing to a mdmg results file
    Returns:
        pandas.DataFrame: A pandas dataframe containing the mdmg results
    """
    mdmg_results = pd.read_csv(file_path, sep=",", index_col=None)
    mdmg_results.rename(columns={"tax_id": "reference"}, inplace=True)
    mdmg_results["Bayesian_significance"] = (
        mdmg_results["Bayesian_D_max"] / mdmg_results["Bayesian_D_max_std"]
    )
    return mdmg_results
